# coding: utf-8
"""
Translator for samba
"""
from eoleaaf.config import basedn, smbstructure
from optparse import OptionParser
from samba.samdb import SamDB
from samba.auth import system_session
import samba.getopt as options
import ldb
import time
from eoleaaf.util import SimpleMongoConnectionWrapper, MongoDBGenerator, modif_fields
from eoleaaf.ihm import print_time


DEBUG_NUM = 100


def erase_db():
    """**CAUTION** use with care : erases all the samba users.
    use it at your own risks...
    """
    parse = OptionParser()
    lp = options.SambaOptions(parse).get_loadparm()
    creds = options.CredentialsOptions(parse).get_credentials(lp)
    samdb = SamDB(session_info=system_session(), credentials=creds, lp=lp)
    dn = samdb.domain_dn()
    #samdb.erase_users_computers(dn)
    try:
        res = samdb.search(base=dn, scope=ldb.SCOPE_SUBTREE, attrs=[],
                  expression="(|(objectclass=group)(objectclass=user)(objectclass=computer)(objectCategory=CN=Organizational-Unit,CN=Schema,CN=Configuration,{}))".format(dn))
    except ldb.LdbError, (errno, _):
        if errno == ldb.ERR_NO_SUCH_OBJECT:
            # Ignore no such object errors
            pass
        else:
            raise
    try:
        for msg in res:
            if not "Domain " in str(msg):
                samdb.delete(msg.dn, ["relax:0"])
    except ldb.LdbError, (errno, _):
        if errno != ldb.ERR_NO_SUCH_OBJECT:
            # Ignore no such object errors
            raise


def verify_indexed():
    parse = OptionParser()
    lp = options.SambaOptions(parse).get_loadparm()
    lp.set("dsdb:schema update allowed", "yes")
    creds = options.CredentialsOptions(parse).get_credentials(lp)
    samdb = SamDB(session_info=system_session(), credentials=creds, lp=lp)
    indexed = False
    for res in samdb.search(base="@INDEXLIST", scope=ldb.SCOPE_BASE, attrs=['@IDXATTR']):
        for i in res['@IDXATTR']:
            if i == 'info':
                indexed = True
    if not indexed:
        modified_attributes = """
dn: CN=Comment,CN=Schema,CN=Configuration,{}
changetype: modify
replace: searchFlags
searchFlags: 1""".format(samdb.domain_dn())
        samdb.modify_ldif(modified_attributes)



class SmbCancel(Exception):
    pass

class SmbImportError(Exception):
    pass


def gen_user_displayname(surname, givenname):
    """
    Génération de l'attribut `displayName`.
    Le `displayName` se construit de la manière suivante :
    Nom et prénom accentués

    :param surname: nom d'usage
    :param givenname: prénom usuel
    """
    if givenname is not None:
        return surname.upper() + ' ' + givenname.title()
    return surname.upper()

class SambaImport(object):
    def __init__(self, log, debug=False, enable_transaction=False):
        self.siren = {}
        self.uai = {}
        #
        self.sqlgen = MongoDBGenerator()
        self._init_connection(enable_transaction)
        self.log = log
        self.debug = debug

    def _init_connection(self, enable_transaction):
        self.db = SimpleMongoConnectionWrapper()
        self.db.connect()
        parse = OptionParser()
        lp = options.SambaOptions(parse).get_loadparm()
        creds = options.CredentialsOptions(parse).get_credentials(lp)
        self.samdb = SamDB(session_info=system_session(), credentials=creds, lp=lp)

        def fake_start():
            pass

        def fake_commit():
            pass

        def fake_cancel():
            self.transaction_cancel()
            raise SmbCancel("transaction cancelled")
        self.enable_transaction = enable_transaction
        self.transaction_start = self.samdb.transaction_start
        self.transaction_commit = self.samdb.transaction_commit
        self.transaction_cancel = self.samdb.transaction_cancel

        self.samdb.transaction_start = fake_start
        self.samdb.transaction_commit = fake_commit
        self.samdb.transaction_cancel = fake_cancel

    def get_typetab_name(self, typ, etabname):
        return etabname + '_' + typ

    def is_present_in_samba(self, groupname, etabname):
        domain_dn = self.calc_group_root_dn(etabname)
        if self.samdb.search(domain_dn, scope=ldb.SCOPE_ONELEVEL,
                    expression=("(&(objectClass=group)(cn={}))".format(groupname)),
                    attrs=["cn"]):
            return True
        return False

    def _create_or_modify_group(self, group, field, etabname):
        tablename, export_type, row = field
        if export_type in ['CREATE', 'DELTACREATE']:
            self.log.debug('create new group {}'.format(group['groupname']))
            self._create_group(group)
        elif export_type in ['DELTAUPDATE', 'DELTAMOVE']:
            if self.enable_transaction:
                raise SmbImportError('cannot update data with an activated transaction')
            self.log.debug('modifying group {}'.format(group['groupname']))
            modified_attributes = """
dn: {0}
changetype: modify
replace: description
description: {1}
""".format(self.calc_group_dn(group['groupname'], etabname), group['description'])
            self.samdb.modify_ldif(modified_attributes)

    def _create_group(self, group):
        self.log.debug('create new group {}'.format(group['groupname']))
        self.samdb.newgroup(**group)

    def _tag_create_or_modified_as_done(self, group):
        tablename, export_type, row = group
        modif_export_type = export_type + "_DONE"
        modif_fields(row, self.sqlgen, action=None, modif_type=modif_export_type, tablename=tablename,
                     fromxml=False, db=self.db, id_column='id', no_bulk=True)
        self.db.commit(tablename, no_bulk=True)

    def create_groups(self, export_type):
        self.log.info('ajout des groupes')
        cnt = 0
        def add_group(grp):
            etabname = grp['etablissement_name']
            groupname = grp['name']
            groupid = grp['oriName']
            groups.append({'kwargs': self.calc_group_kwargs(etabname, groupname, typ, typ[0], groupid),
                           'etabname': etabname,
                           'fields': None
                           })
        transaction_started = False
        # searching for **all** groups
        # informations retrieved for all kind of group
        groups = []
        generic_groups = {'eleve': 'j', 'enseignant': 'k', 'administratif': 'l'}
        for etab in self.db.fetchtable('etablissement', export_type):
            kwargs = self.calc_group_kwargs(etab['ENTStructureUAI'],
                                            etab['ENTStructureUAI'],
                                            etab['ENTStructureNomCourant'],
                                            'e',
                                            etab['ENTStructureJointure'])
            groups.append({'kwargs': kwargs,
                           'etabname': etab['ENTStructureUAI'],
                           'fields': ('etablissement', etab['FieldActionType'], etab)
                          })
            self.create_ou_etab(etab['ENTStructureUAI'])
            for group, type_ref in generic_groups.items():
                groupname = self.get_typetab_name(group, etab['ENTStructureUAI'])
                kwargs = self.calc_group_kwargs(etab['ENTStructureUAI'],
                                                groupname,
                                                etab['ENTStructureNomCourant'],
                                                type_ref,
                                                etab['ENTStructureJointure'])
                groups.append({'kwargs': kwargs,
                               'etabname': etab['ENTStructureUAI'],
                               'fields': None
                              })
        for typ in generic_groups.keys():
            groups.append({'kwargs': self.calc_group_kwargs(None, typ, typ, 'i', typ),
                           'etabname': None,
                           'fields': None
                           })
        for typ in ('classe', 'groupe'):
            for row in self.db.query_groupname(typ):
                if row['_id']:
                    if isinstance(row['_id'], list):
                        for grp in row['_id']:
                            add_group(grp)
                    else:
                        add_group(row['_id'])
        # _____________________________________________________________________
        # creates the groups or modifies the informations about the groups
        for group in groups:
            cnt += 1
            if self.enable_transaction and transaction_started is False:
                self.transaction_start()
                transaction_started = True
            try:
                field = group['fields']
                if field:
                    self._create_or_modify_group(group['kwargs'], field, group['etabname'])
                    if not self.enable_transaction:
                        self._tag_create_or_modified_as_done(field)
                elif not self.is_present_in_samba(group['kwargs']['groupname'], group['etabname']):
                    self._create_group(group['kwargs'])
            except SmbCancel as err:
                if self.debug:
                    self.log.debug(err, exc_info=True)
                if self.enable_transaction:
                    raise err
                else:
                    self.log.error('cannot create or modify the group {}: {}'.format(group['kwargs']['groupname'], str(err)))
                    continue
            except Exception as err:
                if self.debug:
                    self.log.debug(err, exc_info=True)
                if self.enable_transaction:
                    self.transaction_cancel()
                    raise err
                else:
                    self.log.error('cannot create or modify the group {}: {}'.format(group['kwargs']['groupname'], str(err)))
                    continue
            if cnt % DEBUG_NUM == 0:
                self.log.debug("{} {}s traités...".format(cnt, 'groupe'))

        if transaction_started:
            self.transaction_commit()
            for group in groups:
                if group['fields']:
                    self._tag_create_or_modified_as_done(group['fields'])
        self.log.info("TOTAL : {} {}s".format(cnt, 'groupe'))

    def run(self):

        times = [time.time()]
        for export_type in ['CREATE', 'DELTACREATE', 'DELTAUPDATE', 'DELTAMOVE']:
            self.create_groups(export_type)
        for typ in ('eleve', 'enseignant', 'administratif'):
            for export_type in ['CREATE', 'DELTACREATE', 'DELTAUPDATE', 'DELTAMOVE']:
                if export_type == 'CREATE':
                    self.log.info('ajout des {}s'.format(typ))
                elif export_type == 'DELTAUPDATE':
                    self.log.info('mise à jour des {}s'.format(typ))
                self.import_users(typ, export_type)
            print_time(typ, times)

    def _add_or_modify_user(self, export_type, dico, user_dn, etabname):
        if export_type in ["CREATE", "DELTACREATE"]:
            self.log.debug('add new user {}'.format(user_dn))
            self.samdb.newuser(**self.calc_user_kwargs(dico['ENTPersonLogin'], dico['userPassword'], dico['gecos'], etabname))
            registered_groups = frozenset()
        elif export_type in ['DELTAUPDATE', 'DELTAMOVE']:
            self.log.debug('modify user {}'.format(user_dn))
            if export_type == 'DELTAMOVE':
                old_etab = self.db.fetchone(None, None, 'etablissement', '_id', dico['ENTPersonStructRattachsOld'][-1], None)
                if old_etab:
                    old_etab_name = old_etab['ENTStructureUAI']
                    old_user_dn = self.calc_user_dn(str(dico['ENTPersonLogin']), old_etab_name)
                    self.move(old_user_dn, user_dn)
                else:
                    self.log.error("ne peut deplacer l'utilisateur {} dans le nouvel etablissement".format(user_dn))
            modified_attributes = """
dn: {0}
changetype: modify
replace: givenName
givenName: {1}
-
replace: displayName
displayName: {1}
-
replace: gecos
gecos: {1}
""".format(user_dn, dico['gecos'])
            self.samdb.modify_ldif(modified_attributes)
            # **registered_groups**, that is the groups where the user is already in
            registered_groups = set(self.get_groups_by_user(user_dn, etabname))
        return registered_groups

    def import_users(self, typ, export_type):
        cnt = 0
        transaction_started = False
        for dico in self.db.fetchtable(typ, export_type):
            try:
                cnt += 1
                # **expected_groups** means the user is expected to belong to these groups
                expected_groups = set([self.calc_group_dn(typ, None)])
                etab = self.db.fetchone(None, None, 'etablissement', '_id', dico['ENTPersonStructRattach'], None)
                if etab:
                    etab_name = etab['ENTStructureUAI']
                    expected_groups.add(self.calc_group_dn(etab_name, etab_name))
                    expected_groups.add(self.calc_group_dn(self.get_typetab_name(typ, etab_name), etab_name))
                else:
                    etab_name = None
                user_dn = self.calc_user_dn(str(dico['ENTPersonLogin']), etab_name)
                if transaction_started is False:
                    self.transaction_start()
                    transaction_started = True
                registered_groups = self._add_or_modify_user(export_type, dico, user_dn, etab_name)
                # let's add the etablissement at the user's group list
                # if it's a student, we add the student in his classroom
                if typ == 'eleve':
                    expected_groups.add(self.calc_group_dn(dico['ENTEleveClasses']['name'], etab_name))
                    if isinstance(dico['ENTEleveGroupes'], list):
                        for grp in dico['ENTEleveGroupes']:
                            expected_groups.add(self.calc_group_dn(grp['name'], etab_name))
                    else:
                        expected_groups.add(self.calc_group_dn(dico['ENTEleveGroupes']['name'], etab_name))
                # let's add the user to the groups in which he is not already registered
                groups_to_add = expected_groups - registered_groups
                for group in groups_to_add:
                    self.add_member_to_group(group, user_dn)
                # let's remove the user to the groups in which he is not supposed to be registered
                groups_to_del = registered_groups - expected_groups
                for group in groups_to_del:
                    self.del_member_to_group(group, user_dn)
                # if the transaction_mode is not enabled,
                # we have to set the add/modify of the user as done
                if not self.enable_transaction:
                    modif_export_type = export_type + "_DONE"
                    modif_fields(dico, self.sqlgen, action=None, modif_type=modif_export_type, tablename=typ,
                                fromxml=False, db=self.db, id_column='id', no_bulk=True)
                    self.db.commit('user', no_bulk=True)
                if not self.enable_transaction:
                    self.transaction_commit()
                    transaction_started = False
                if cnt % DEBUG_NUM == 0:
                    self.log.info("{} {}s traités...".format(cnt, typ))
            except SmbCancel as err:
                transaction_started = False
                if self.debug:
                    self.log.debug(err, exc_info=True)
                if self.enable_transaction:
                    raise err
                else:
                    self.log.error('cannot create or modify the user {}: {}'.format(dico['ENTPersonLogin'], str(err)))
                    continue
            except Exception as err:
                if self.debug:
                    self.log.debug(err, exc_info=True)
                if transaction_started:
                    self.transaction_cancel()
                    transaction_started = False
                if self.enable_transaction:
                    raise err
                else:
                    self.log.error('cannot create or modify the user {}: {}'.format(dico['ENTPersonLogin'], str(err)))
                    continue
        if transaction_started:
            self.transaction_commit()
            for row in self.db.fetchtable(typ, export_type):
                modif_export_type = export_type + "_DONE"
                modif_fields(row, self.sqlgen, action=None, modif_type=modif_export_type, tablename=typ,
                            fromxml=False, db=self.db, id_column='id')
            self.db.commit('user')
        self.log.info("TOTAL : {} {}s".format(cnt, typ))

    def add_member_to_group(self, group_dn, member):
        self.log.debug('add user {} to group {}'.format(member, group_dn))
        ldif = """dn: {}
changetype: modify
add: member
member: {}
""".format(group_dn, member)
        self.samdb.modify_ldif(ldif)

    def del_member_to_group(self, group_dn, member):
        self.log.debug('delete user {} to group {}'.format(member, group_dn))
        ldif = """dn: {}
changetype: modify
delete: member
member: {}
""".format(group_dn, member)
        self.samdb.modify_ldif(ldif)

    def get_groups_by_user(self, username, etabname):
        user = username.split(',')[0]
        userfilter = "(&(%s)(objectCategory=%s,%s))" % (user, "CN=Person,CN=Schema,CN=Configuration", self.samdb.domain_dn())
        for res in self.samdb.search(base=self.calc_user_root_dn(etabname), scope=ldb.SCOPE_SUBTREE,
                expression=userfilter, attrs=['memberOf']):
            if 'memberOf' in res:
                for member in res['memberOf']:
                    yield member


class SambaTranslator(SambaImport):

    def calc_user_kwargs(self, username, password, gecos, etabname):
        gecos = gecos.encode('utf-8')
        return dict(username=str(username),
                    password=password,
                    userou=self.calc_user_ou(etabname),
                    givenname=str(gecos),
                    gecos=str(gecos),
                    useusernameascn=True)

    def calc_group_cn(self, groupname):
        return groupname

    def calc_group_desc(self, grp, typ):
        if typ == 'classe':
            return 'Classe de {}'.format(grp)
        elif typ == 'groupe':
            return 'Groupe de {}'.format(grp)
        else:
            return 'Groupe {}'.format(typ.encode('utf-8'))

    def calc_group_kwargs(self, etabname, groupname, typ, type_ref, ref):
        """group calculation

        :param etabname: name of the etablissement as created in the AD (usually UIA)
        :param groupname: name of the group as created in the AD
        :param typ: enables us tu build the group's description
        :param type_ref: only one char in order to define the type of the group.
                         This char can be set to

                         - classe: 'c'
                         - etablissement: 'e'
                         - group: 'g'
                         - eleve enseignant and admin global: 'i'
                         - eleve in an etablissement: 'j'
                         - enseignant in an etablissement: 'k'
                         - administratif in an etablissement: 'l'
        :param ref: id of the group in mongodb
        """
        return dict(groupname=str(self.calc_group_cn(groupname)),
                    groupou=self.calc_group_ou(etabname),
                    description=self.calc_group_desc(groupname, typ),
                    notes=str(type_ref + ref))


# _____________________________________________________________________________
# Type of samba used (singleton determined by the configuration file)
class SambaToplevel(SambaTranslator):
    """Standard usage of the samba
    """
    toplevel_ou = "CN=Users"

    def calc_group_ou(self, etabname):
        return self.toplevel_ou

    def calc_group_root_dn(self, etabname):
        return "{},{}".format(self.calc_group_ou(etabname), basedn)

    def calc_group_dn(self, groupname, etabname):
        return "CN={},{}".format(groupname, self.calc_group_root_dn(etabname))

    def calc_user_ou(self, etabname):
        return self.calc_group_ou(etabname)

    def calc_user_root_dn(self, etabname):
        return "{},{}".format(self.calc_user_ou(etabname), basedn)

    def calc_user_dn(self, value, etabname):
        return "CN={},{}".format(value, self.calc_user_root_dn(etabname))

    def create_ou_etab(self, etabname):
        pass

    def move(self, old_user_dn, user_dn):
        pass


class SambaSplit(SambaToplevel):
    """Splitted by OU (organisation unit)
    """
    def calc_group_ou(self, etabname):
        if etabname is None:
            return self.toplevel_ou
        else:
            return "OU={}".format(etabname)

    def create_ou_etab(self, etabname):
        oufilter = "(&(%s)(objectCategory=%s,%s))" % ("OU={}".format(etabname), "CN=Organizational-Unit,CN=Schema,CN=Configuration", self.samdb.domain_dn())
        if not self.samdb.search(base=self.samdb.domain_dn(), scope=ldb.SCOPE_SUBTREE, attrs=[],
                expression=oufilter):
            ldb_create_ou = """dn: OU={},{}
changetype: add
objectClass: top
objectClass: organizationalunit
description: my first ou
""".format(etabname, self.samdb.domain_dn())
            self.samdb.modify_ldif(ldb_create_ou)

    def move(self, old_user_dn, user_dn):
        self.samdb.rename(old_user_dn, user_dn)


# _____________________________________________________________________________
# singleton definition
toplevel_factory = {
        'toplevel': SambaToplevel,
        'split': SambaSplit
        }

SambaExport = toplevel_factory[smbstructure]
# _____________________________________________________________________________
